#include <cmath>
#include <vector>
#include <deque>
#include <utility>
#include <stdio.h>

template<typename T>
class TemporalAntialias
{
    double target_interval;
    double current_time;
    typedef std::pair<T, double> elem_t;
    typedef std::deque<elem_t> rep_t;
    rep_t buffer;
    double minim;
    double antiblur_factor;
public:
    TemporalAntialias(double interval)
        : target_interval(interval), current_time(0.0), minim(0.01),
          antiblur_factor(4.5)
    {
    }
    void Feed(double time, const T& frame)
    {
        buffer.push_back(elem_t(frame, time));
    }
    void FeedAndEat(double time, T& frame)
    {
        /*fprintf(stderr, "received frame at %g\n", time);*/
        buffer.push_back(elem_t(frame, time));
        /*buffer.push_back(elem_t(T(), time));
        buffer.back().first.swap(frame);*/
    }
    /* Avail() returns true if there are enough samples
     *         to fully render the next frame
     */
    bool Avail() const
    {
        if(buffer.empty()) return false;
        const double next_time = current_time + target_interval;
        return buffer.back().second >= next_time;
    }

    void SetAntiBlur(double v) { antiblur_factor = v; }
    void SetMinim(double m) { minim = m; }

    /* Get() renders the next frame and pops all samples
     * that will not be needed after that
     */
    template<typename Functor>
    unsigned Get(Functor& func, bool get_only_latest = false)
    {
        unsigned result = 0;
        const double next_time = current_time + target_interval;
        if(get_only_latest)
        {
            typename rep_t::iterator latest = buffer.begin();
            for(typename rep_t::iterator i = buffer.begin(); i != buffer.end(); ++i)
            {
                if(i->second >= next_time) break;
                latest = i;
            }
            func(latest->first, 128);
            result += 1;
        }
        else
        {
            double density = 0.0;
            /* Collect the frames to sample */
            std::vector<double> weight;
            for(typename rep_t::iterator i = buffer.begin(); i != buffer.end(); ++i)
            {
                if(i->second >= next_time) break;
                double distance = i->second - current_time;
                if(distance < 0) distance = -distance;
                double this_frame_weight = 1.0 - distance/target_interval;
                this_frame_weight = std::pow(this_frame_weight, antiblur_factor);
                weight.push_back(this_frame_weight);
                if(this_frame_weight < minim) continue;
                density += this_frame_weight;
            }
            /* Sample them with weights. Total weight for frame = 1.0 */
            size_t counter = 0;
            for(typename rep_t::iterator i = buffer.begin(); i != buffer.end(); ++i)
            {
                if(i->second >= next_time) break;
                double this_frame_weight = weight[counter++];
                /*fprintf(stderr, "got weight %g, minim = %g\n",
                    this_frame_weight, minim);*/
                if(this_frame_weight < minim) continue;
                this_frame_weight /= density;
                int  intweight = this_frame_weight * 4194304.0 + 0.5;
                fprintf(stderr, "%12.4g: Sampling frame at %12.4g for power %g (%d)\n",
                    current_time,
                    i->second, this_frame_weight, intweight);
                func(i->first, intweight);
                ++result;
            }
            if(!counter)
            {
                fprintf(stderr, "%12.4g: Sampling nothing? From buffer size=%u\n", current_time, (unsigned)buffer.size());
            }
        }
        /* Pop all samples that will not be needed by frame after this */
        while(! buffer.empty()
             && buffer.front().second <= current_time)
        {
            buffer.front().first.donewith();
            buffer.pop_front();
        }
        /* Advance to next frame */
        current_time = next_time;
        return result;
    }
};
